/*
 * SNAC - Server-based list management
 * Mitz Pettel, 2001
 *
 * based on: SNAC - Buddy (Contact) list management
 * Copyright (C) 2001 Barnaby Gray <barnaby@beedesign.co.uk>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef SNAC_SBL_H
#define SNAC_SBL_H

#include <string>
#include <cstdio>
#include <list>
#include <vector>

#include <libicq2000/SNAC-base.h>
#include <libicq2000/Contact.h>
#include <libicq2000/ContactList.h>
#include <libicq2000/UserInfoBlock.h>

namespace ICQ2000 {

  // Server-based list stuff (Family 0x0013)
  const unsigned short SNAC_SBL_Error = 0x0001;
  const unsigned short SNAC_SBL_Request_Params = 0x0002;
  const unsigned short SNAC_SBL_List_Params = 0x0003;
  const unsigned short SNAC_SBL_Import_List = 0x0004;
  const unsigned short SNAC_SBL_Request_List = 0x0005;
  const unsigned short SNAC_SBL_List_From_Server = 0x0006;
  const unsigned short SNAC_SBL_Activate_List = 0x0007;

  const unsigned short SNAC_SBL_Add_Item = 0x0008;
  const unsigned short SNAC_SBL_Update_Group = 0x0009;
  const unsigned short SNAC_SBL_Remove_Item = 0x000a;
  const unsigned short SNAC_SBL_Edit_Ack = 0x000e;
  const unsigned short SNAC_SBL_Sync_Result = 0x000f;
  const unsigned short SNAC_SBL_Edit_Start = 0x0011;
  const unsigned short SNAC_SBL_Edit_End = 0x0012;
  const unsigned short SNAC_SBL_Auth_Grant = 0x0014;
  const unsigned short SNAC_SBL_Auth_Granted = 0x0015;
  const unsigned short SNAC_SBL_Remove_Me = 0x0016;
  const unsigned short SNAC_SBL_Auth_Request_Send = 0x0018;
  const unsigned short SNAC_SBL_Auth_Request_Ack = 0x0019;
  const unsigned short SNAC_SBL_Auth_Reply_Send = 0x001a;
  const unsigned short SNAC_SBL_Auth_Reply_Ack = 0x001b;
  const unsigned short SNAC_SBL_Auth_Added = 0x001c;
  

  // ----------------- Server-based Lists (Family 0x0013) SNACs -----------

  class SBLFamilySNAC : virtual public SNAC {
   public:
    unsigned short Family() const { return SNAC_FAM_SBL; }
  };

  class RequestSBLSNAC : public SBLFamilySNAC, public OutSNAC {
    
   protected:
    void OutputBody(Buffer& b) const;

   public:
    RequestSBLSNAC();

    unsigned short Subtype() const { return SNAC_SBL_Import_List; }
  };
  
  class ActivateSBLSNAC : public SBLFamilySNAC, public OutSNAC {
    
   protected:
    void OutputBody(Buffer& b) const;

   public:
    ActivateSBLSNAC();

    unsigned short Subtype() const { return SNAC_SBL_Activate_List; }
  };
  
  class SBLParamsSNAC : public SBLFamilySNAC, public OutSNAC {
    
   protected:
    void OutputBody(Buffer& b) const;

   public:
    SBLParamsSNAC();

    unsigned short Subtype() const { return SNAC_SBL_Request_Params; }
  };
  
  class SBLListSNAC : public SBLFamilySNAC, public InSNAC {
   private:
    ContactList m_contacts;
    std::map<std::string, unsigned short> m_groups;
     
   protected:
    void ParseBody(Buffer& b);

   public:
    SBLListSNAC();
    
    const ContactList& getContactList() const { return m_contacts; }
    const std::map<std::string, unsigned short> & getGroupMap() const { return m_groups; }
    unsigned short Subtype() const { return SNAC_SBL_List_From_Server; }
  };

  class SBLParamsListSNAC : public SBLFamilySNAC, public InSNAC {
   private:
//    ContactList m_contacts;
     
   protected:
    void ParseBody(Buffer& b);

   public:
    SBLParamsListSNAC();
    
//    const ContactList& getContactList() const { return m_contacts; }
    unsigned short Subtype() const { return SNAC_SBL_List_Params; }
  };

  class SBCLSyncDoneSNAC : public SBLFamilySNAC, public OutSNAC {
   private:
    unsigned int m_ts;
    unsigned short m_count;
     
   protected:
    void OutputBody(Buffer& b) const;

   public:
    SBCLSyncDoneSNAC(unsigned int, unsigned short);
    
    unsigned short Subtype() const { return SNAC_SBL_Request_List; }
  };

  class SBL_SSI_Edit_Start : public SBLFamilySNAC, public OutSNAC {
   private:
     
   protected:
    void OutputBody(Buffer& b) const {};

   public:
    SBL_SSI_Edit_Start() {};
    
    unsigned short Subtype() const { return SNAC_SBL_Edit_Start; }
  };

  class SBL_SSI_Edit_End : public SBLFamilySNAC, public OutSNAC {
   private:
     
   protected:
    void OutputBody(Buffer& b) const {};

   public:
    SBL_SSI_Edit_End() {};
    
    unsigned short Subtype() const { return SNAC_SBL_Edit_End; }
  };

  
  class SBL_SSI_Add_Item : public SBLFamilySNAC, public OutSNAC {
   private:
     std::string m_uin;
     unsigned short m_groupid, m_itemid;
     bool m_req;

   protected:
    void OutputBody(Buffer& b) const;

   public:
    SBL_SSI_Add_Item(std::string uin, unsigned short gid, unsigned short itid) : m_uin(uin), m_groupid(gid), m_itemid(itid), m_req(false) {};
    void setAuthReq(bool req) { m_req = req; }
    unsigned short Subtype() const { return SNAC_SBL_Add_Item; }
  };

  class SBL_SSI_Add_Group : public SBLFamilySNAC, public OutSNAC {
   private:
     std::string m_group;
     unsigned short m_groupid;

   protected:
    void OutputBody(Buffer& b) const;

   public:
    SBL_SSI_Add_Group(std::string name, unsigned short gid) : m_group(name), m_groupid(gid) {};
    
    unsigned short Subtype() const { return SNAC_SBL_Add_Item; }
  };

  class SBL_SSI_Remove_Item : public SBLFamilySNAC, public OutSNAC {
   private:
     std::string m_uin;
     unsigned short m_groupid, m_itemid;
     
   protected:
    void OutputBody(Buffer& b) const;

   public:
    SBL_SSI_Remove_Item(std::string uin, unsigned short gid, unsigned short itid) : m_uin(uin), m_groupid(gid), m_itemid(itid) {};
    
    unsigned short Subtype() const { return SNAC_SBL_Remove_Item; }
  };

  class SBL_SSI_Edit_Ack : public SBLFamilySNAC, public InSNAC {
   private:
     std::vector<unsigned short> m_codes;
   protected:
    void ParseBody(Buffer& b);

   public:
    SBL_SSI_Edit_Ack() {};
    
    unsigned short Subtype() const { return SNAC_SBL_Edit_Ack; }
     std::vector<unsigned short> const getCodes() { return m_codes; };
  };

  class SBL_Auth_Request_Send : public SBLFamilySNAC, public OutSNAC {
   private:
     unsigned int m_uin;
     std::string m_reason;
     
   protected:
    void OutputBody(Buffer& b) const;

   public:
    SBL_Auth_Request_Send(unsigned int uin, std::string reason="Please add me to your contact list!") : m_uin(uin), m_reason(reason) {};
    
    unsigned short Subtype() const { return SNAC_SBL_Auth_Request_Send; }
  };

  class SBL_Auth_Reply_Ack : public SBLFamilySNAC, public InSNAC {
   private:
     std::string m_uin;
     bool m_accepted;
     std::string m_reason;
   protected:
    void ParseBody(Buffer& b);

   public:
    SBL_Auth_Reply_Ack() {};
    
    unsigned short Subtype() const { return SNAC_SBL_Auth_Reply_Ack; }
    std::string const getUin() { return m_uin; };
    bool const getAccepted() { return m_accepted; };
    std::string const getReason() { return m_reason; };
  };

  class SBL_FutureAuth_Grant : public SBLFamilySNAC, public OutSNAC {
   private:
     unsigned int m_uin;
     std::string m_reason;
     
   protected:
    void OutputBody(Buffer& b) const;

   public:
	   SBL_FutureAuth_Grant(unsigned int uin, std::string reason) : m_uin(uin), m_reason(reason) {};
    
    unsigned short Subtype() const { return SNAC_SBL_Auth_Grant; }
  };

  class SBL_RemoveMe_SNAC : public SBLFamilySNAC, public OutSNAC {
   private:
     std::string m_uin;
     
   protected:
    void OutputBody(Buffer& b) const;

   public:
	   SBL_RemoveMe_SNAC(std::string uin) : m_uin(uin) {};
    
    unsigned short Subtype() const { return SNAC_SBL_Remove_Me; }
  };

  
}

#endif
