/*
 * Capabilities
 *
 * Copyright (C) 2002 Barnaby Gray <barnaby@beedesign.co.uk>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 */

#include <libicq2000/Capabilities.h>
#include <stdio.h>
#include <string.h>
#include <sstream>

using std::set;

namespace ICQ2000 {

  /*
   * Capability blocks.  (Thanks libfaim for figuring out these - many
   * are only relevant to AIM, but included for completeness)
   *
   * These are CLSIDs. They should actually be of the form:
   *
   * {0x0946134b, 0x4c7f, 0x11d1,
   *  {0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},
   *
   * But, eh.
   */

  const struct Capabilities::Block Capabilities::caps[] = {
    /*
     * Chat is oddball.
     */
    {Chat,
     {0x74, 0x8f, 0x24, 0x20, 0x62, 0x87, 0x11, 0xd1,
      0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},

    /*
     * These are mostly in order.
     */
    {Voice,
     {0x09, 0x46, 0x13, 0x41, 0x4c, 0x7f, 0x11, 0xd1,
      0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},

    {SendFile,
     {0x09, 0x46, 0x13, 0x43, 0x4c, 0x7f, 0x11, 0xd1,
      0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},

    /*
     * Advertised by the EveryBuddy client.
     */
    {ICQa,
     {0x09, 0x46, 0x13, 0x44, 0x4c, 0x7f, 0x11, 0xd1,
      0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},

    {IMImage,
     {0x09, 0x46, 0x13, 0x45, 0x4c, 0x7f, 0x11, 0xd1,
      0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},

    {BuddyIcon,
     {0x09, 0x46, 0x13, 0x46, 0x4c, 0x7f, 0x11, 0xd1,
      0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},

    {SaveStocks,
     {0x09, 0x46, 0x13, 0x47, 0x4c, 0x7f, 0x11, 0xd1,
      0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},

    {GetFile,
     {0x09, 0x46, 0x13, 0x48, 0x4c, 0x7f, 0x11, 0xd1,
      0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},

    {ICQServerRelay,
     {0x09, 0x46, 0x13, 0x49, 0x4c, 0x7f, 0x11, 0xd1,
      0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},

    /*
     * Indeed, there are two of these.  The former appears to be correct,
     * but in some versions of winaim, the second one is set.  Either they
     * forgot to fix endianness, or they made a typo. It really doesn't
     * matter which.
     */
    {Games,
     {0x09, 0x46, 0x13, 0x4a, 0x4c, 0x7f, 0x11, 0xd1,
      0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},
    {Games2,
     {0x09, 0x46, 0x13, 0x4a, 0x4c, 0x7f, 0x11, 0xd1,
      0x22, 0x82, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},

    {SendBuddyList,
     {0x09, 0x46, 0x13, 0x4b, 0x4c, 0x7f, 0x11, 0xd1,
      0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},

    {ICQUTF8,
     {0x09, 0x46, 0x13, 0x4e, 0x4c, 0x7f, 0x11, 0xd1,
      0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},

    {ICQRTF,
     {0x97, 0xb1, 0x27, 0x51, 0x24, 0x3c, 0x43, 0x34,
      0xad, 0x22, 0xd6, 0xab, 0xf7, 0x3f, 0x14, 0x92}},

    {ICQUnknown,
     {0x2e, 0x7a, 0x64, 0x75, 0xfa, 0xdf, 0x4d, 0xc8,
      0x88, 0x6f, 0xea, 0x35, 0x95, 0xfd, 0xb6, 0xdf}},

    {Empty,
     {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}},

    {TrillianCrypt,
     {0xf2, 0xe7, 0xc7, 0xf4, 0xfe, 0xad, 0x4d, 0xfb,
      0xb2, 0x35, 0x36, 0x79, 0x8b, 0xdf, 0x00, 0x00}},

    {APInfo,
     {0xaa, 0x4a, 0x32, 0xb5, 0xf8, 0x84, 0x48, 0xc6,
      0xa3, 0xd7, 0x8c, 0x50, 0x97, 0x19, 0xfd, 0x5b}},

    {MacICQ,
     {0xDD, 0x16, 0xF2, 0x02, 0x84, 0xE6, 0x11, 0xD4,
      0x90, 0xDB, 0x00, 0x10, 0x4B, 0x9B, 0x4B, 0x7D}},

    {AIMInteroperate,
     {0x09, 0x46, 0x13, 0x4d, 0x4c, 0x7f, 0x11, 0xd1,
      0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},

    {AIM_FileTransfer,
     {0x09, 0x46, 0x13, 0x4c, 0x4c, 0x7f, 0x11, 0xd1,
      0x82, 0x22, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00}},

    {Typing_Notifications,
     {0x56, 0x3F, 0xC8, 0x09, 0x0B, 0x6F, 0x41, 0xBD,
      0x9F, 0x79, 0x42, 0x26, 0x09, 0xDF, 0xA2, 0xF3}},
    {Client_QIP,
     {0x56, 0x3F, 0xC8, 0x09, 0x0B, 0x6F, 0x41, 0x51,
      0x49, 0x50, 0x20, 0x32, 0x30, 0x30, 0x35, 0x61}},
    {QIP_ProtectMSG,
     {0xD3, 0xD4, 0x53, 0x19, 0x8B, 0x32, 0x40, 0x3B,
      0xAC, 0xC7, 0xD1, 0xA9, 0xE2, 0xB5, 0x81, 0x3E}},
    {xTraz,
     {0x1A, 0x09, 0x3C, 0x6C, 0xD7, 0xFD, 0x4E, 0xC5,
      0x9D, 0x51, 0xA6, 0x47, 0x4E, 0x34, 0xF5, 0xA0}},
// xStatus caps from QIP forums and distribution:
  // QIP's status: XStatus(Angry)
      {xStatus01, {0x01, 0xD8, 0xD7, 0xEE, 0xAC, 0x3B, 0x49, 0x2A, 0xA5, 0x8D, 0xD3, 0xD8, 0x77, 0xE6, 0x6B, 0x92}},
  // QIP's status: XStatus(Duck)
      {xStatus02, {0x5A, 0x58, 0x1E, 0xA1, 0xE5, 0x80, 0x43, 0x0C, 0xA0, 0x6F, 0x61, 0x22, 0x98, 0xB7, 0xE4, 0xC7}},
  // QIP's status: XStatus(Tired)
      {xStatus03, {0x83, 0xC9, 0xB7, 0x8E, 0x77, 0xE7, 0x43, 0x78, 0xB2, 0xC5, 0xFB, 0x6C, 0xFC, 0xC3, 0x5B, 0xEC}},
  // QIP's status: XStatus(Party)
      {xStatus04, {0xE6, 0x01, 0xE4, 0x1C, 0x33, 0x73, 0x4B, 0xD1, 0xBC, 0x06, 0x81, 0x1D, 0x6C, 0x32, 0x3D, 0x81}},
  // QIP's status: XStatus(Beer)
      {xStatus05, {0x8C, 0x50, 0xDB, 0xAE, 0x81, 0xED, 0x47, 0x86, 0xAC, 0xCA, 0x16, 0xCC, 0x32, 0x13, 0xC7, 0xB7}},
  // QIP's status: XStatus(Thinking)
      {xStatus06, {0x3F, 0xB0, 0xBD, 0x36, 0xAF, 0x3B, 0x4A, 0x60, 0x9E, 0xEF, 0xCF, 0x19, 0x0F, 0x6A, 0x5A, 0x7F}},
  // QIP's status: XStatus(Eating)
      {xStatus07, {0xF8, 0xE8, 0xD7, 0xB2, 0x82, 0xC4, 0x41, 0x42, 0x90, 0xF8, 0x10, 0xC6, 0xCE, 0x0A, 0x89, 0xA6}},
  // QIP's status: XStatus(TV)
      {xStatus08, {0x80, 0x53, 0x7D, 0xE2, 0xA4, 0x67, 0x4A, 0x76, 0xB3, 0x54, 0x6D, 0xFD, 0x07, 0x5F, 0x5E, 0xC6}},
  // QIP's status: XStatus(Friends)
      {xStatus09, {0xF1, 0x8A, 0xB5, 0x2E, 0xDC, 0x57, 0x49, 0x1D, 0x99, 0xDC, 0x64, 0x44, 0x50, 0x24, 0x57, 0xAF}},
  // QIP's status: XStatus(Coffee)
      {xStatus10, {0x1B, 0x78, 0xAE, 0x31, 0xFA, 0x0B, 0x4D, 0x38, 0x93, 0xD1, 0x99, 0x7E, 0xEE, 0xAF, 0xB2, 0x18}},
  // QIP's status: XStatus(Music)
      {xStatus11, {0x61, 0xBE, 0xE0, 0xDD, 0x8B, 0xDD, 0x47, 0x5D, 0x8D, 0xEE, 0x5F, 0x4B, 0xAA, 0xCF, 0x19, 0xA7}},
  // QIP's status: XStatus(Business)
      {xStatus12, {0x48, 0x8E, 0x14, 0x89, 0x8A, 0xCA, 0x4A, 0x08, 0x82, 0xAA, 0x77, 0xCE, 0x7A, 0x16, 0x52, 0x08}},
  // QIP's status: XStatus(Camera)
      {xStatus13, {0x10, 0x7A, 0x9A, 0x18, 0x12, 0x32, 0x4D, 0xA4, 0xB6, 0xCD, 0x08, 0x79, 0xDB, 0x78, 0x0F, 0x09}},
  // QIP's status: XStatus(Funny)
      {xStatus14, {0x6F, 0x49, 0x30, 0x98, 0x4F, 0x7C, 0x4A, 0xFF, 0xA2, 0x76, 0x34, 0xA0, 0x3B, 0xCE, 0xAE, 0xA7}},
  // QIP's status: XStatus(Phone)
      {xStatus15, {0x12, 0x92, 0xE5, 0x50, 0x1B, 0x64, 0x4F, 0x66, 0xB2, 0x06, 0xB2, 0x9A, 0xF3, 0x78, 0xE4, 0x8D}},
  // QIP's status: XStatus(Games)
      {xStatus16, {0xD4, 0xA6, 0x11, 0xD0, 0x8F, 0x01, 0x4E, 0xC0, 0x92, 0x23, 0xC5, 0xB6, 0xBE, 0xC6, 0xCC, 0xF0}},
  // QIP's status: XStatus(College)
      {xStatus17, {0x60, 0x9D, 0x52, 0xF8, 0xA2, 0x9A, 0x49, 0xA6, 0xB2, 0xA0, 0x25, 0x24, 0xC5, 0xE9, 0xD2, 0x60}},
  // QIP's status: XStatus(Shopping)
      {xStatus18, {0x63, 0x62, 0x73, 0x37, 0xA0, 0x3F, 0x49, 0xFF, 0x80, 0xE5, 0xF7, 0x09, 0xCD, 0xE0, 0xA4, 0xEE}},
  // QIP's status: XStatus(Sick)
      {xStatus19, {0x1F, 0x7A, 0x40, 0x71, 0xBF, 0x3B, 0x4E, 0x60, 0xBC, 0x32, 0x4C, 0x57, 0x87, 0xB0, 0x4C, 0xF1}},
  // QIP's status: XStatus(Sleeping)
      {xStatus20, {0x78, 0x5E, 0x8C, 0x48, 0x40, 0xD3, 0x4C, 0x65, 0x88, 0x6F, 0x04, 0xCF, 0x3F, 0x3F, 0x43, 0xDF}},
  // QIP's status: XStatus(Surfing)
      {xStatus21, {0xA6, 0xED, 0x55, 0x7E, 0x6B, 0xF7, 0x44, 0xD4, 0xA5, 0xD4, 0xD2, 0xE7, 0xD9, 0x5C, 0xE8, 0x1F}},
  // QIP's status: XStatus(@)
      {xStatus22, {0x12, 0xD0, 0x7E, 0x3E, 0xF8, 0x85, 0x48, 0x9E, 0x8E, 0x97, 0xA7, 0x2A, 0x65, 0x51, 0xE5, 0x8D}},
  // QIP's status: XStatus(Engineering)
      {xStatus23, {0xBA, 0x74, 0xDB, 0x3E, 0x9E, 0x24, 0x43, 0x4B, 0x87, 0xB6, 0x2F, 0x6B, 0x8D, 0xFE, 0xE5, 0x0F}},
  // QIP's status: XStatus(Typing)
      {xStatus24, {0x63, 0x4F, 0x6B, 0xD8, 0xAD, 0xD2, 0x4A, 0xA1, 0xAA, 0xB9, 0x11, 0x5B, 0xC2, 0x6D, 0x05, 0xA1}},
  // QIP's status: XStatus(China1)
      {xStatus25, {0x2C, 0xE0, 0xE4, 0xE5, 0x7C, 0x64, 0x43, 0x70, 0x9C, 0x3A, 0x7A, 0x1C, 0xE8, 0x78, 0xA7, 0xDC}},
  // QIP's status: XStatus(China2)
      {xStatus26, {0x10, 0x11, 0x17, 0xC9, 0xA3, 0xB0, 0x40, 0xF9, 0x81, 0xAC, 0x49, 0xE1, 0x59, 0xFB, 0xD5, 0xD4}},
  // QIP's status: XStatus(China3)
      {xStatus27, {0x16, 0x0C, 0x60, 0xBB, 0xDD, 0x44, 0x43, 0xF3, 0x91, 0x40, 0x05, 0x0F, 0x00, 0xE6, 0xC0, 0x09}},
  // QIP's status: XStatus(China4)
      {xStatus28, {0x64, 0x43, 0xC6, 0xAF, 0x22, 0x60, 0x45, 0x17, 0xB5, 0x8C, 0xD7, 0xDF, 0x8E, 0x29, 0x03, 0x52}},
  // QIP's status: XStatus(China5)
      {xStatus29, {0x16, 0xF5, 0xB7, 0x6F, 0xA9, 0xD2, 0x40, 0x35, 0x8C, 0xC5, 0xC0, 0x84, 0x70, 0x3C, 0x98, 0xFA}},
  // QIP's status: XStatus(De1)
      {xStatus30, {0x63, 0x14, 0x36, 0xFF, 0x3F, 0x8A, 0x40, 0xD0, 0xA5, 0xCB, 0x7B, 0x66, 0xE0, 0x51, 0xB3, 0x64}},
  // QIP's status: XStatus(De2)
      {xStatus31, {0xB7, 0x08, 0x67, 0xF5, 0x38, 0x25, 0x43, 0x27, 0xA1, 0xFF, 0xCF, 0x4C, 0xC1, 0x93, 0x97, 0x97}},
  // QIP's status: XStatus(De3)
      {xStatus32, {0xDD, 0xCF, 0x0E, 0xA9, 0x71, 0x95, 0x40, 0x48, 0xA9, 0xC6, 0x41, 0x32, 0x06, 0xD6, 0xF2, 0x80}},
  // QIP's status: XStatus(RuSearch)
      {xStatus33, {0xD4, 0xE2, 0xB0, 0xBA, 0x33, 0x4E, 0x4F, 0xA5, 0x98, 0xD0, 0x11, 0x7D, 0xBF, 0x4D, 0x3C, 0xC8}},
  // QIP's status: XStatus(RuLove)
      {xStatus34, {0xCD, 0x56, 0x43, 0xA2, 0xC9, 0x4C, 0x47, 0x24, 0xB5, 0x2C, 0xDC, 0x01, 0x24, 0xA1, 0xD0, 0xCD}},
  // QIP's status: XStatus(RuJournal)
      {xStatus35, {0x00, 0x72, 0xD9, 0x08, 0x4A, 0xD1, 0x43, 0xDD, 0x91, 0x99, 0x6F, 0x02, 0x69, 0x66, 0x02, 0x6F}},

      {Client_JIT,{0x4A, 0x49, 0x54, 0x20, 0x76, 0x2E, 0x31, 0x2E, 0x78, 0x2E, 0x78, 0x00, 0x00, 0x00, 0x00, 0x00}}
  };


  Capabilities::Capabilities()
  { }

  void Capabilities::default_icq2000_capabilities()
  {
    clear();
    set_capability_flag( ICQa );
    set_capability_flag( ICQServerRelay );
// MITZ for UTF8 support
    set_capability_flag( ICQUTF8 );
  }

  void Capabilities::default_jit_capabilities()
  {
    clear();
    set_capability_flag( ICQa );
    set_capability_flag( ICQServerRelay );
    set_capability_flag( ICQUTF8 );
    set_capability_flag( Client_JIT );
    set_capability_flag( xTraz );
  }

  void Capabilities::default_icq2002_capabilities()
  {
    clear();
    set_capability_flag( ICQa );
    set_capability_flag( ICQServerRelay );
    set_capability_flag( ICQRTF );
    set_capability_flag( ICQUTF8 );
  }

  void Capabilities::clear()
  {
    m_flags.clear();
  }

  void Capabilities::set_capability_flag(Flag f)
  {
    m_flags.insert(f);
  }

  void Capabilities::clear_capability_flag(Flag f)
  {
    m_flags.erase(f);
  }

  bool Capabilities::has_capability_flag(Flag f) const
  {
    return (m_flags.count(f) > 0);
  }
  
  std::string Capabilities::guess_client_version() const
  {
	  if(has_capability_flag(Client_QIP)) {
		  return "QIP 2005";
	  } else 
	  if(has_capability_flag(Client_JIT)) {
		  return "JIT 1.2.0";
	  } else {
		  return "Unknown client";
	  }
  }

  void Capabilities::Parse(Buffer& b, unsigned short len)
  {
    int i = len / sizeof_cap;
    unsigned char cap[sizeof_cap];
    for (int j = 0; j < i; ++j) {
      b.Unpack(cap, sizeof_cap);

      // search for capability in list
      bool setflag = false;
      for (unsigned int k = 0; k < sizeof(caps) / sizeof(Block); ++k)
    	  if ( memcmp( caps[k].data, cap, sizeof_cap ) == 0 ) {
    		  fprintf(stderr, "Cap %d found\n", k);
    		  set_capability_flag( caps[k].flag );
    		  setflag = true;
    		  break;
    	  }
      if(!setflag) {
    	  char capstr[39];
    	  sprintf( capstr, "{%02X%02X%02X%02X-%02X%02X-%02X%02X-%02X%02X-%02X%02X%02X%02X%02X%02X}", cap[0], cap[1], cap[2], cap[3],
    				cap[4], cap[5], cap[6], cap[7], cap[8], cap[9],
    				cap[10], cap[11], cap[12], cap[13], cap[14], cap[15] );
    	  fprintf(stderr, "Unknown capability received: %s\n", capstr);
      }
      // might be nice to catch unknown capabilities
    }

    // any remainder (there shouldn't be any, but..)
    b.advance( len - i * sizeof_cap );
  }

  void Capabilities::ParseString(Buffer& b, unsigned short len)
  {
    int i = len;
	unsigned l;
	bool inside = false;
    unsigned char cap[sizeof_cap];
    while ( i>0 ) {
	  unsigned char c, d;
	  b >> c;
	  i--;
	  switch (c)
	  {
	    case '{':
		  inside = true;
		  l = 0;
		  break;
		case '}':
		  if ( inside )
		  {
		    inside = false;
			if ( l==sizeof_cap )
			{
			  // search for capability in list
			  for (unsigned int k = 0; k < sizeof(caps) / sizeof(Block); ++k)
				if ( memcmp( caps[k].data, cap, sizeof_cap ) == 0 ) {
				  set_capability_flag( caps[k].flag );
				  break;
				}
			}
		  }
		  break;
		default:
		  if ( inside )
		  {
		    if ( (c>='0' && c<='9' || c>='A' && c<='F' || c>='a' && c<='f') && i>0 )
			{
			  if ( l==sizeof_cap )
			    inside = false;
			  else
			  {
			    b >> d;
			    i--;
				  if ( d>='0' && d<='9' || d>='A' && d<='F' || d>='a' && d<='f' )
			  	    cap[l++] = ( ( c>='a' ? c+10-'a' : ( c>='A' ? c+10-'A' : c-'0' ) )<<4 )
							  + ( d>='a' ? d+10-'a' : ( d>='A' ? d+10-'A' : d-'0' ) );
			  }
			}
		  }
		  break;
	  }
    }
  }

  void Capabilities::OutputString( Buffer& b ) const
  {
    for (set<Flag>::const_iterator curr = m_flags.begin();
	 curr != m_flags.end(); ++curr)
      for (unsigned int i = 0; i < sizeof(caps) / sizeof(Block); ++i)
	if ( caps[i].flag == *curr ) {
	  char capstr[39];
	  const unsigned char *data = caps[i].data;
	  sprintf( capstr, "{%02X%02X%02X%02X-%02X%02X-%02X%02X-%02X%02X-%02X%02X%02X%02X%02X%02X}", data[0], data[1], data[2], data[3],
				data[4], data[5], data[6], data[7], data[8], data[9],
				data[10], data[11], data[12], data[13], data[14], data[15] );
	  b.Pack( (const unsigned char *)capstr, 38 );
	  break;
	}
  }

  void Capabilities::Output(Buffer& b) const
  {
    for (set<Flag>::const_iterator curr = m_flags.begin();
	 curr != m_flags.end(); ++curr)
      for (unsigned int i = 0; i < sizeof(caps) / sizeof(Block); ++i)
	if ( caps[i].flag == *curr ) {
	  b.Pack( caps[i].data, sizeof_cap );
	  break;
	}

  }

  unsigned short Capabilities::get_length() const
  {
    return sizeof_cap * m_flags.size();
  }

  /*
   * This can be done properly based on the actual capability now - as
   * opposed to the dodgy method based on length that was used before.
   */
  bool Capabilities::get_accept_adv_msgs() const
  {
    return (has_capability_flag(ICQa) && has_capability_flag(ICQServerRelay));
  }

  std::string Capabilities::getMood() const
  {
	  for(int i=xStatus01; i<=xStatus35; ++i) {
		  if(has_capability_flag((Flag)i)) {
			  std::stringstream out;
			  out << "InXMood" <<((int)i-(int)xStatus01);
			  return out.str();
		  }
	  }
	  return "";
  }
}

