/*
 * SNAC - New UIN services
 *
 * Copyright (C) 2001 Barnaby Gray <barnaby@beedesign.co.uk>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 */

#include <libicq2000/SNAC-UIN.h>
#include <libicq2000/TLV.h>

#ifdef _FREEBSD7
typedef int _Bool;
#endif
#include <mhash.h>

#define AIM_MD5_STRING "AOL Instant Messenger (SM)"

#include <libicq2000/buffer.h>

using std::string;

namespace ICQ2000 {

  // -------------- New UIN (0x0017) Family -----------------------

  UINRequestSNAC::UINRequestSNAC(const string& p)
    : m_password(p) { }

  void UINRequestSNAC::OutputBody(Buffer& b) const{
    b<<(unsigned int)0x00010039;
    b<<(unsigned int)0x0;
    b<<(unsigned int)0x28000300;
    b<<(unsigned int)0x0;
    b<<(unsigned int)0x0;
    b<<(unsigned int)0x624e0000;
    b<<(unsigned int)0x624e0000;
    b<<(unsigned int)0x0;
    b<<(unsigned int)0x0;
    b<<(unsigned int)0x0;
    b<<(unsigned int)0x0;
    b.setLittleEndian();
    b.PackUint16StringNull(m_password);
    b.setBigEndian();
    b<<(unsigned int)0x624e0000;
    b<<(unsigned int)0x0000d601;
  }

  UINResponseSNAC::UINResponseSNAC() { }

  void UINResponseSNAC::ParseBody(Buffer& b){
    b.advance(46);
    b.setLittleEndian();
    b >> m_uin;
    b.advance(10);
  }
  
  UINRequestErrorSNAC::UINRequestErrorSNAC() { }
  
  void UINRequestErrorSNAC::ParseBody(Buffer& b){
    b.advance(32);
  }

  AuthCookieRequestSNAC::AuthCookieRequestSNAC(const std::string screen)
  { 
    m_uin = screen;
  }

  void AuthCookieRequestSNAC::OutputBody(Buffer& b) const{
	std::cout << "AuthCookieRequestSNAC" << std::endl;
	b<<ScreenNameTLV(m_uin);
	
    b<<(unsigned short)0x4b; // unknown (command?)
    b<<(unsigned short)0x0; // icon id
  }

  AuthCookieResponseSNAC::AuthCookieResponseSNAC() : m_data("") { }

  void AuthCookieResponseSNAC::ParseBody(Buffer& b){
	std::cout << "AuthCookieResponseSNAC" << std::endl;
    b.dump(std::cout);
    b >> m_data;
  }
  
  std::string AuthCookieResponseSNAC::get_key() {
    return m_data;
  }
  

  AuthRequestSNAC::AuthRequestSNAC(const std::string screen, const std::string key, const std::string password)
  { 
    m_uin = screen;
    m_authkey = key;
    m_passwd = password;
  }

  void AuthRequestSNAC::OutputBody(Buffer& b) const{
	std::cout << "AuthRequestSNAC" << std::endl;

    b << ScreenNameTLV(m_uin);
    // Digest password calculation 
    MHASH td, td2;
    unsigned char hash[16];
    td = mhash_init(MHASH_MD5);
    td2 = mhash_init(MHASH_MD5);
    
    if(td == MHASH_FAILED || td2 == MHASH_FAILED) {
	// Something goes wrong
	return;
    }
    mhash(td2, m_passwd.c_str(), m_passwd.length());
    mhash_deinit(td2, hash);
    
    mhash(td, m_authkey.c_str(), m_authkey.length());
    mhash(td, hash, 16);
    mhash(td, AIM_MD5_STRING, strlen(AIM_MD5_STRING));
    
    mhash_deinit(td, hash);
    
    b << (unsigned short) 0x25
      << (unsigned short) 0x10;
    
    for(int i=0; i<16; ++i)
	b << (unsigned char) hash[i];
    b << (unsigned short) 0x4c;
    b << (unsigned short) 0x00;
    
    b << (unsigned short) 0x00a2
      << (unsigned short) 0x0002
      << (unsigned short) 0x0005;
      
    b << (unsigned short) 0x00a3
      << (unsigned short) 0x0002
      << (unsigned short) 0x0005;
      
    b << (unsigned short) 0x00a4
      << (unsigned short) 0x0002
      << (unsigned short) 0x0000;

    b << (unsigned short) 0x00a5
      << (unsigned short) 0x0002
      << (unsigned short) 0x17f2;
      
    b << ClientProfileTLV("ICQ Client")
      << ClientVersionMajorTLV(6)
      << ClientVersionMinorTLV(5)
      << ClientICQNumberTLV(0)
      << ClientBuildMajorTLV(0x0068)
      << ClientTypeTLV(0x0068)
      << ClientBuildMinorTLV(0x7553)
      << LanguageTLV("ru")
      << CountryCodeTLV("ru");
    
    b << (unsigned short) 0x004a
      << (unsigned short) 0x0001
      << (unsigned char) 0x01;
  }

  AuthResponseSNAC::AuthResponseSNAC() { 
	  m_host = ""; 
	  m_port = 0;
	  m_error = 0;
	  m_cookie = "";
  }

  void AuthResponseSNAC::ParseBody(Buffer& b){
	std::cout << "AuthResponseSNAC" << std::endl;
    TLVList tlvlist;
    tlvlist.Parse(b, TLV_ParseMode_Channel04, (unsigned int)-1);
    if(!tlvlist.exists(TLV_ErrorCode)) {
    	m_error = 0;
    	RedirectTLV *r = static_cast<RedirectTLV*>(tlvlist[TLV_Redirect]);
    	m_host = r->getHost();
    	m_port = r->getPort();
    	
    	CookieTLV *c = static_cast<CookieTLV*>(tlvlist[TLV_Cookie]);
    	m_cookie.assign((const char*)c->Value(), c->Length());
    } else {
    	ErrorCodeTLV *e = static_cast<ErrorCodeTLV*>(tlvlist[TLV_ErrorCode]);
    	m_error = e->Value();
	if(tlvlist.exists(TLV_ErrorText)) {
	    ErrorTextTLV *e = static_cast<ErrorTextTLV*>(tlvlist[TLV_ErrorText]);
	    m_host = e->Value();
	}
    }
  }
  
}
